/*
 * Copyright (c) 2004 Changwoo Ryu
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */
#include "mbprintf.h"
#include <stdlib.h>
#include <sys/types.h>
#include <regex.h>
#include <wchar.h>
#include <assert.h>

#define RE_FLAGS	"\\([-+ #']\\+\\)\\?"
#define RE_WIDTH	"\\([-+ #']\\?[0-9]\\+\\|\\*\\)\\?"
#define RE_PRECISION	"\\(\\.\\([0-9]\\+\\|\\*\\)\\)\\?"
#define RE_MODIFIER	"\\(hh\\|h\\|j\\|l\\|L\\|ll\\|q\\|t\\|z\\|Z\\)\\?"
#define RE_CONVERSION	"[%diouxXfeEgGaAcCsSpnm]"

#define RE "\\(%" RE_FLAGS "\\)" RE_WIDTH RE_PRECISION "\\(" RE_MODIFIER RE_CONVERSION "\\)"

#define SE_PREFIX 1
#define SE_SUFFIX 6
#define SE_FLAGS 2
#define SE_WIDTH 3
#define SE_PRECISION 5
#define SE_MODIFIER 7

#define NSUB 8

#ifdef DEBUG
#define TRACE(format, ...) fprintf(stderr, format, ##__VA_ARGS__)
#else
#define TRACE(format, ...)
#endif


static int
mbswidth(const char *str)
{
  size_t len;
  wchar_t *wcs;
  int width;

  len = mbstowcs(0, str, 0);
  wcs = malloc(sizeof(wchar_t) * len);

  len = mbstowcs(wcs, str, len);
  width = wcswidth(wcs, len);

  free(wcs);
  return width;
}

static int
mbscrop(const char *str, int maxwidth, int *realwidth)
{
  mbstate_t state;
  int totallength = 0;
  int totalwidth = 0;

  memset(&state, 0, sizeof(state));
  while(1) {
    wchar_t wc;
    int w;
    int length = mbrtowc(&wc, str, strlen(str), &state);
    if (length <= 0) break;

    w = wcwidth(wc);

    if ((totalwidth + w) > maxwidth)
      break;
    totalwidth += w;
    totallength += length;
    str += length;
  }
  *realwidth = totalwidth;
  return totallength;
}

static void
parsewidth(const char *format, regmatch_t *match,
	   int *has_width, int *width)
{
  char buf[16];
  *has_width = (match[SE_WIDTH].rm_so >= 0);
  if (*has_width)
    {
      /* parse the width in the format */
      strncpy(buf, format + match[SE_WIDTH].rm_so,
	      match[SE_WIDTH].rm_eo - match[SE_WIDTH].rm_so);
      buf[match[SE_WIDTH].rm_eo - match[SE_WIDTH].rm_so] = '\0';
      *width = atoi(buf);
    }
}

static void
parseprecision(const char *format, regmatch_t *match,
	       int *has_precision, int *precision)
{
  char buf[16];
  *has_precision = (match[SE_PRECISION].rm_so >= 0);
  if (*has_precision)
    {
      /* parse the precision in the format */
      strncpy(buf, format + match[SE_PRECISION].rm_so,
	      match[SE_PRECISION].rm_eo - match[SE_PRECISION].rm_so);
      buf[match[SE_PRECISION].rm_eo - match[SE_PRECISION].rm_so] = '\0';
      *precision = atoi(buf);
    }
}


static int
fprintmbstring(FILE *stream, const char *format, int nformat,
	       regmatch_t *match,
	       int has_width, int width, int has_precision, int precision,
	       char *arg_str)
{
  char buf[256];
  int extrapadding = 0;
  int len;

  TRACE("arg: \"%s\"\n", arg_str);

  if (has_width)
    {
      int realwidth = mbswidth(arg_str);
      int padding = (width - realwidth);

      if (padding > 0)
	width = strlen(arg_str) + padding; /* need more length */

      TRACE("width: %d\n", width);
    }

  has_precision = (match[SE_PRECISION].rm_so >= 0);
  if (has_precision)
    {
      int realwidth;
      len = mbscrop(arg_str, precision, &realwidth);

      if (realwidth < precision)
	extrapadding = precision - realwidth;
      else
	extrapadding = 0;

      if (len > precision)
	precision = len;

      TRACE("precision: %d\n", precision);
      TRACE("extrapadding: %d\n", extrapadding);
    }

  /* prefix */
  len = match[SE_PREFIX].rm_eo;
  strncpy(buf, format, match[SE_PREFIX].rm_eo);
  len = match[SE_PREFIX].rm_eo;
  buf[len] = '\0';

  /* the modified width/precision, if any */
  if (has_width)
    {
      sprintf(buf + len, "%d", width);
      len += strlen(buf+len);
    }

  if (has_precision)
    {
      sprintf(buf + len, ".%d", precision);
      len += strlen(buf+len);
    }

  /* suffix */
  strncpy(buf + len, format + match[SE_SUFFIX].rm_so,
	  match[0].rm_eo - match[SE_SUFFIX].rm_so);
  len += match[0].rm_eo - match[SE_SUFFIX].rm_so;
  buf[len] = '\0';

  if (has_precision & extrapadding)
    {
      while(extrapadding--)
	buf[len++] = ' ';
      buf[len] = '\0';
    }

  TRACE("new format: \"%s\"\n", buf);

  return fprintf(stream, buf, arg_str);
}


#define PRINTFORMAT(argtype,promtype) \
do {								\
  argtype arg;							\
  strncpy(buf, format, match[0].rm_eo);				\
  buf[match[0].rm_eo] = '\0';					\
  arg = va_arg(ap, promtype);					\
  if (has_width && has_precision)				\
    total += fprintf(stream, buf, width, precision, arg);	\
  else if (has_width)						\
    total += fprintf(stream, buf, width, precision, arg);	\
  else if (has_precision)					\
    total += fprintf(stream, buf, precision, arg);		\
  else								\
    total += fprintf(stream, buf, arg);				\
} while(0)

#define MODIFIER0 \
((match[SE_MODIFIER].rm_so < 0) ? '\0' : format[match[SE_MODIFIER].rm_so])
#define MODIFIER1 \
((match[SE_MODIFIER].rm_so < 0 || match[SE_MODIFIER].rm_so == match[SE_MODIFIER].rm_eo) ? '\0' : format[match[SE_MODIFIER].rm_so+1])


int
vfmbprintf (FILE *stream, const char *format, va_list ap)
{
  int ret;
  regex_t re;
  regmatch_t match[NSUB];
  char buf[256];					\
  int total = 0;

  TRACE("\"%s\"\n", format);

  ret = regcomp(&re, RE, 0);
  assert (ret == 0);


  while (1)
    {
      char convtype;
      int has_width = 0, width;
      int has_precision = 0, precision;

      ret = regexec(&re, format, NSUB, match, 0);
      if (ret != 0)
	{
	  total += fprintf(stream, format);
	  break;
	}

      convtype = format[match[0].rm_eo - 1];

      if (match[SE_WIDTH].rm_so >= 0 &&
	  format[match[SE_WIDTH].rm_so] == '*')
	{
	  has_width = 1;
	  width = va_arg(ap, int);
	}

      if (match[SE_PRECISION].rm_so >= 0 &&
	  format[match[SE_PRECISION].rm_so] == '*')
	{
	  has_precision = 1;
	  precision = va_arg(ap, int);
	}

      switch (convtype)
	{
	case 'd':
	case 'i':
	case 'o':
	case 'u':
	case 'x':
	case 'X':
	  switch (MODIFIER0)
	    {
	    case 'j':
	      PRINTFORMAT(int, int);
	      break;
	    case 'l':
	      if (MODIFIER1 == 'l')
		PRINTFORMAT(long long int, long long int);
	      else
		PRINTFORMAT(long int, long int);
	      break;
	    case 'L':
	    case 'q':
	      PRINTFORMAT(long int, long int);
	      break;
#if 0
	    case 't':
	      PRINTFORMAT(ptrdiff_t, ptrdiff_t);
	      break;
#endif
	    case 'z':
	    case 'Z':
	      PRINTFORMAT(size_t, size_t);
	      break;
	    default:
	      PRINTFORMAT(int, int);
	      break;
	    }
	  break;
	case 'f':
	case 'e':
	case 'E':
	case 'g':
	case 'G':
	case 'a':
	case 'A':
	  if (MODIFIER0 == 'L')
	    PRINTFORMAT(long double, long double);
	  else
	    PRINTFORMAT(double, double);
	  break;	  
	case 'c':
	  if (MODIFIER0 == 'l')
	    PRINTFORMAT(wint_t, wint_t);
	  else
	    PRINTFORMAT(char, int);
	  break;
	case 'S':
	  PRINTFORMAT(wchar_t*, wchar_t*);
	  break;
	case 'p':
	  PRINTFORMAT(void*, void*);
	  break;
	case 'n':
	case 'm':
	case '%':
	  strncpy(buf, format, match[0].rm_so);
	  buf[match[0].rm_so] = '\0';
	  total += fprintf(stream, buf);
	  break;
	case 's':
	  if (match[SE_WIDTH].rm_so >= 0 && !has_width)
	    parsewidth(format, match, &has_width, &width);
	  if (match[SE_PRECISION].rm_so >= 0 && !has_precision)
	    parseprecision(format, match, &has_precision, &precision);
	  if (has_width || has_precision)
	    {
	      char *arg_str = va_arg(ap, char*);
	      total += fprintmbstring(stream, format, match[0].rm_eo, match,
				      has_width, width,
				      has_precision, precision,
				      arg_str);
	    }
	  else
	    PRINTFORMAT(char*, char*);
	  break;
	}





      format += match[0].rm_eo;
    }

  regfree(&re);

  return total;
}

int
vmbprintf (const char *format, va_list ap)
{
  return vfmbprintf (stdout, format, ap);
}

int
fmbprintf (FILE *stream, const char *format, ...)
{
  va_list args;
  int ret;

  va_start(args, format);
  ret = vfmbprintf (stream, format, args);
  va_end(args);
  return ret;
}

int
mbprintf (const char *format, ...)
{
  va_list args;
  int ret;

  va_start(args, format);
  ret = vfmbprintf (stdout, format, args);
  va_end(args);
  return ret;
}

#if 0 /* TODO */

int
smbprintf (const char *format, ...)
{
  va_list args;
  int ret;

  va_start(args, format);
  ret = vsmbprintf (stdout, format, args);
  va_end(args);
  return ret;
}

int
snmbprintf (FILE *stream, const char *format, ...)
{
  va_list args;
  int ret;

  va_start(args, format);
  ret = vsnmbprintf (stream, format, args);
  va_end(args);
  return ret;
}

int
vsnmbprintf (FILE *stream, const char *format, ...)
{
  va_list args;
  int ret;

  va_start(args, format);
  ret = vsmbprintf (stream, format, args);
  va_end(args);
  return ret;
}

#endif
